/*
 * DataFile.java
 *
 * Created on 24  2004 ., 14:50
 */

package com.fn.BikersLog;


import java.io.File;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;


/**
 *
 * @author  asso
 */
public class DataFile {

    public static final short CURRENT_MAJOR_VERSION = 1;
    public static final short CURRENT_MINOR_VERSION = 6;
    
    private static final int BUFFER_SIZE = 102400;
    private static final int DATA_OFFSET = 12;
    private static final int MAX_FREE_BLOCKS = 50;
    private static final double MAX_FREE_RATIO = 2;
    
    private class Entry implements Comparable {
        String name;
        long offset;
        long size;
        byte[] inMemory;
        
        public Entry(RandomAccessFile raf) throws IOException {
            name = raf.readUTF();
            offset = raf.readLong();
            size = raf.readLong();
            inMemory = null;
        }
        
        public Entry(String name, byte[] data) {
            this.name = name;
            inMemory = data;
            size = data.length;
        }
        
        public String getName() {
            return name;
        }
        
        public int compareTo(Object o) {
            Entry e = (Entry)o;
            if (offset < e.offset)
                return -1;
            if (offset > e.offset)
                return 1;
            return 0;
        }
        
        public void write(RandomAccessFile rndFile) throws IOException {
            rndFile.writeUTF(name);
            rndFile.writeLong(offset);
            rndFile.writeLong(size);
        }
        
        public boolean equals(Object obj) {
            return ((this == obj) || name.equals(((Entry)obj).name));
        }
        
    }
    
    private class FreeBlock implements Comparable {
        long offset;
        long size;
        
        public FreeBlock(long o, long s) {
            offset = o;
            size = s;
        }
        
        public int compareTo(Object o) {
            FreeBlock b = (FreeBlock)o;
            if (offset < b.offset)
                return -1;
            if (offset > b.offset)
                return 1;
            return 0;
        }
    }
    
    private RandomAccessFile rndFile = null;
    private Map entries = new HashMap();
    private List dataBlocks = new LinkedList();
    private List freeBlocks = new LinkedList();
    private long lastOffset = 0;
    private short majorVersion = CURRENT_MAJOR_VERSION;
    private short minorVersion = CURRENT_MINOR_VERSION;
    private File file = null;
    private int revision = 0;
    private boolean unnamed = false;
    
    public DataFile() {
    }
    
    /** Read data file */
    public DataFile(File file) throws Exception {
        this.file = file;
        rndFile = new RandomAccessFile(file, "r");
        byte sign[] = new byte[4];
        rndFile.readFully(sign);
        
        if ((0x3C == sign[0]) && (0x3F == sign[1]) && (0x78 == sign[2])
                 && (0x6D == sign[3]))
            loadLegacyFile();
        else {
            if ((0x53 != sign[0]) || (0x50 != sign[1]) || (0x54 != sign[2])
                     || (0x46 != sign[3]))
                throw new Exception(I18n.getMsg("dataFile.invalidSignature"));
            majorVersion = rndFile.readShort();
            minorVersion = rndFile.readShort();
            if (majorVersion > 1)
                throw new Exception(I18n.getMsg("dataFile.incompatibleVersion"));
            revision = rndFile.readInt();
            readDirectory();
        }
    }
    
    private void loadLegacyFile() throws IOException {
        majorVersion = 1;
        minorVersion = 5;
        byte data[] = new byte[(int)rndFile.length()];
        rndFile.seek(0);
        rndFile.readFully(data);
        rndFile.close();
        rndFile = null;
        setData("log.xml", data);
    }
    
    private void readDirectory() throws IOException {        
        long len = rndFile.length();
        rndFile.seek(len - 12);
        int entriesCnt = rndFile.readInt();
        long dirOffset = rndFile.readLong();
        lastOffset = dirOffset;
        
        rndFile.seek(dirOffset);
        for (int i = 0; i < entriesCnt; i++) {
            Entry e = new Entry(rndFile);
            dataBlocks.add(e);
            entries.put(e.getName(), e);
        }
    }

    private void buildFreeBlocksList() {
        freeBlocks.clear();
        lastOffset = DATA_OFFSET;
        for (Iterator i = dataBlocks.iterator(); i.hasNext(); ) {
            Entry e = (Entry)i.next();
            if (e.offset > lastOffset)
                freeBlocks.add(new FreeBlock(lastOffset, e.offset - lastOffset));
            lastOffset = e.offset + e.size;
        }
    }

    private void moveBlock(long offset, long oldOffset, long size, 
            byte[] buffer) throws IOException
    {
        while (size > 0) {
            int toProcess = size > BUFFER_SIZE ? BUFFER_SIZE : (int)size;
            rndFile.seek(oldOffset);
            rndFile.readFully(buffer, 0, toProcess);
            rndFile.seek(offset);
            rndFile.write(buffer, 0, toProcess);
            offset += toProcess;
            oldOffset += toProcess;
            size -= toProcess;
        }
    }
    
    private void relocateBlocks() throws IOException {
        byte buffer[] = null;

        long offset = DATA_OFFSET;
        for (Iterator i = freeBlocks.iterator(); i.hasNext(); ) {
            Entry e = (Entry)i.next();
            if (e.offset > offset) {
                if (null == buffer)
                    buffer = new byte[BUFFER_SIZE];
                moveBlock(offset, e.offset, e.size, buffer);
                e.offset = offset;
                offset += e.size;
            }
        }
        
        freeBlocks.clear();
        lastOffset = offset;
    }
    
    private void writeHeader() throws IOException {
        byte sign[] = { 0x53, 0x50, 0x54, 0x46 };
        rndFile.seek(0);
        rndFile.write(sign);
        rndFile.writeShort(1);
        rndFile.writeShort(6);
        rndFile.writeInt(revision);
    }
    
    private long allocateSpace(long size) {
        if (0 == freeBlocks.size()) {
            long res = lastOffset;
            lastOffset += size;
            return res;
        }
        
        FreeBlock bestBlock = null;
        for (Iterator i = freeBlocks.iterator(); i.hasNext(); ) {
            FreeBlock fb = (FreeBlock)i.next();
            if (fb.size == size) {
                i.remove();
                return fb.offset;
            }
            
            if (fb.size > size) {
                if (null == bestBlock)
                    bestBlock = fb;
                else
                    if (bestBlock.size > fb.size)
                        bestBlock = fb;
            }
        }
        
        if (null != bestBlock) {
            long res = bestBlock.offset;
            bestBlock.offset += size;
            bestBlock.size -= size;
            return res;
        } else {
            long res = lastOffset;
            lastOffset += size;
            return res;
        }
    }
    
    private void writeUnsavedBlocks() throws IOException {
        for (Iterator i = entries.values().iterator(); i.hasNext(); ) {
            Entry e = (Entry)i.next();
            if (null != e.inMemory) {
                long offset = allocateSpace(e.size);
                rndFile.seek(offset);
                rndFile.write(e.inMemory);
                e.offset = offset;
                e.inMemory = null;
                dataBlocks.add(e);
            }
        }
    }
    
    private void checkRevision() throws Exception {
        rndFile.seek(8);
        int r = rndFile.readInt();
        if (r != revision)
            throw new Exception("dataFile.invalidRevision");
    }
    
    private boolean isOptimizationRequiered() {
        if (dataBlocks.size() < 2)
            return false;
        
        if (freeBlocks.size() >= MAX_FREE_BLOCKS)
            return true;

        long freeSize = 0;
        for (Iterator i = freeBlocks.iterator(); i.hasNext(); )
            freeSize += ((FreeBlock)i.next()).size;
        
        long usedSize = 0;
        for (Iterator i = entries.values().iterator(); i.hasNext(); )
            usedSize += ((Entry)i.next()).size;
        
        if ((double)freeSize / (double)usedSize > MAX_FREE_RATIO)
            return true;
            
        return false;
    }
    
    private void preloadData() throws IOException {
        for (Iterator i = dataBlocks.iterator(); i.hasNext(); ) {
            Entry e = (Entry)i;
            rndFile.seek(e.offset);
            e.inMemory = new byte[(int)e.size];
            rndFile.readFully(e.inMemory);
            i.remove();
        }
    }
    
    public void writeAs(File newFileName) throws Exception {
        unnamed = false;
        
        file = newFileName;
        if (rndFile != null) {
            if (dataBlocks.size() > 0)
                preloadData();
            rndFile.close();
            rndFile = null;
        }
        revision = 0;
        
        write();
    }
    
    private void writeDirectory() throws IOException {
        rndFile.seek(lastOffset);
        for (Iterator i = dataBlocks.iterator(); i.hasNext(); )
            ((Entry)i.next()).write(rndFile);
        rndFile.writeInt(dataBlocks.size());
        rndFile.writeLong(lastOffset);
        rndFile.setLength(rndFile.getFilePointer());
    }
    
    public void write() throws Exception {
        // reopen file for read and write
        if (null != rndFile) {
            if (0 != dataBlocks.size())
                checkRevision();
            rndFile.close();
        }
        
        if (Config.getCreateBackup())
            Utils.saveBackupFile(file);
        
        RandomAccessFile f;
        try {
            f = new RandomAccessFile(file, "rw");
            rndFile = f;
        } catch (IOException e) {
            if (null != rndFile)
                rndFile = new RandomAccessFile(file, "r");
            throw e;
        }

        revision++;
        writeHeader();
        
        Collections.sort(dataBlocks);
        if (isOptimizationRequiered()) {
            relocateBlocks();
        } else
            buildFreeBlocksList();
        
        writeUnsavedBlocks();
        writeDirectory();

        freeBlocks.clear();
        // reopen file for read only
        rndFile = new RandomAccessFile(file, "r");
    }
    
    public short getMajorVersion() {
        return majorVersion;
    }
    
    public short getMinorVersion() {
        return minorVersion;
    }
    
    public File getFile() {
        if (unnamed)
            return null;
        return file;
    }
    
    public InputStream getResourseInput(String name) throws Exception {
        Entry e = (Entry)entries.get(name);
        if (null == e)
            throw new Exception(I18n.getMsg("dataFile.resourceNotFound"));
        if (null != e.inMemory)
            return new java.io.ByteArrayInputStream(e.inMemory);
        else
            return new OffsetInputStream(rndFile, e.offset, e.size);
    }
    
    public OutputStream getResourseOutput(String name) {
        return new ResourceOuputStream(this, name);
    }
    
    public OutputStream getResourseOutput(String name, int size) {
        return new ResourceOuputStream(this, name, size);
    }
    
    public void setData(String name, byte[] data) {
        Entry e = (Entry)entries.get(name);
        if (null == e) {
            entries.put(name, new Entry(name, data));
        } else {
            if (null == e.inMemory)
                dataBlocks.remove(e);
            e.inMemory = data;
            e.size = data.length;
        }
    }
    
    public byte[] getData(String name) throws Exception {
        Entry e = (Entry)entries.get(name);
        if (null == e)
            throw new Exception(I18n.getMsg("dataFile.resourceNotFound"));
        if (null != e.inMemory)
            return e.inMemory;
        else {
            rndFile.seek(e.offset);
            byte buf[] = new byte[(int)e.size];
            rndFile.readFully(buf);
            return buf;
        }
    }
    
    public void remove(String name) {
        Entry e = (Entry)entries.get(name);
        if (null != e) {
            entries.remove(e);
            dataBlocks.remove(e);
        }
    }
    
    public void setUnnamed() {
        unnamed = true;
    }
}
